#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

__docformat__ = 'restructuredtext'

from datetime import datetime

from elasticsearch_dsl import Q, Search
from pyramid.threadlocal import get_current_registry
from pyramid_es import get_client

from pyams_catalog.query import CatalogResultSet
from pyams_content.features.search import ISearchFolder
from pyams_content.shared.view.interfaces import IViewQuery, IViewQueryEsParamsExtension, IViewQueryFilterExtension, \
    IViewUserQuery, IWfView, RELEVANCE_ORDER
from pyams_utils.adapter import ContextAdapter, adapter_config
from pyams_utils.list import unique_iter
from pyams_utils.request import check_request
from pyams_utils.timezone import tztime
from pyams_workflow.interfaces import IWorkflow


@adapter_config(name='es', context=IWfView, provides=IViewQuery)
class EsViewQuery(ContextAdapter):
    """View query for Elasticsearch"""

    def get_params(self, context):
        view = self.context
        registry = get_current_registry()
        # check publication dates
        now = tztime(datetime.utcnow())
        params = Q('range', **{'workflow.effective_date': {'lte': now}})
        # check workflow states
        wf_params = []
        for workflow in registry.getAllUtilitiesRegisteredFor(IWorkflow):
            wf_params.extend(workflow.published_states)
        params &= Q('terms', **{'workflow.status': wf_params})
        # check custom extensions
        do_search = True
        for name, adapter in sorted(registry.getAdapters((view,), IViewQueryEsParamsExtension),
                                    key=lambda x: x[1].weight):
            for new_params in adapter.get_es_params(context):
                if isinstance(new_params, tuple):
                    new_params, do_search = new_params
                if new_params:
                    params &= new_params
            if not do_search:
                break
        # activate search
        if do_search:
            params &= Q('bool', must=Q('range', **{'workflow.push_end_date': {'gte': now}})) | \
                      Q('bool', must_not=Q('exists', **{'field': 'workflow.push_end_date'}))
            # check content path
            content_path = view.get_content_path(context)
            if content_path is not None:
                params &= Q('term', **{'parent_ids': content_path})
            # check content types
            content_types = view.get_content_types(context)
            if content_types:
                params &= Q('terms', **{'content_type': content_types})
            # check data types
            data_types = view.get_data_types(context)
            if data_types:
                params &= Q('terms', **{'data_type': data_types})
        return params

    def get_results(self, context, sort_index, reverse, limit):
        request = check_request()
        registry = request.registry
        client = get_client(request)
        params = self.get_params(context)
        search = Search(using=client.es, index=client.index) \
            .query(params) \
            .source(['internal_id'])
        if sort_index != RELEVANCE_ORDER:
            search = search.sort({
                'workflow.{0}'.format(sort_index): {
                    'order': 'desc' if reverse else 'asc',
                    'unmapped_type': 'date'
                }
            })
        if limit:
            search = search[:limit]
        else:
            search = search[:999]
        items = CatalogResultSet([result.internal_id for result in search])
        for name, adapter in sorted(registry.getAdapters((self.context,), IViewQueryFilterExtension),
                                    key=lambda x: x[1].weight):
            items = adapter.filter(context, items)
        return unique_iter(items), search.count()


@adapter_config(name='es', context=ISearchFolder, provides=IViewQuery)
class EsSearchFolderQuery(EsViewQuery):
    """Search folder view query for Elasticsearch"""

    def get_params(self, context):
        params = super(EsSearchFolderQuery, self).get_params(context)
        request = check_request()
        registry = request.registry
        for name, adapter in registry.getAdapters((self,), IViewUserQuery):
            for user_param in adapter.get_user_params(request):
                params &= user_param
        return params


@adapter_config(name='user_search', context=EsSearchFolderQuery, provides=IViewUserQuery)
class EsSearchFolderUserQuery(ContextAdapter):
    """Search folder user query for Elasticsearch"""

    @staticmethod
    def get_user_params(request):
        params = request.params
        fulltext = params.get('user_search')
        if fulltext:
            yield Q('simple_query_string', **{'query': fulltext})
